@@ -7,21 +7,22 @@ set -x
 
 if [ "$1" = "--in-docker" ]; then
     # Runs in docker, performs the actual build.
+    shift
 
-    apk add bash curl gcc g++ make musl-dev bsd-compat-headers git python perl
+    apk add bash curl gcc g++ make musl-dev bsd-compat-headers git python3 perl
 
     git clone /v /librdkafka
 
     cd /librdkafka
-    ./configure --install-deps --disable-gssapi --disable-lz4-ext --enable-static
+    ./configure --install-deps --disable-gssapi --disable-lz4-ext --enable-static $*
     make -j
     examples/rdkafka_example -X builtin.features
-    make -C tests run_local
+    CI=true make -C tests run_local_quick
 
     # Create a tarball in artifacts/
     cd src
     ldd librdkafka.so.1
-    tar cvzf /v/artifacts/alpine-librdkafka.tgz librdkafka.so.1
+    tar cvzf /v/artifacts/alpine-librdkafka.tgz librdkafka.so.1 librdkafka-static.a rdkafka-static.pc
     cd ../..
 
 else
@@ -33,5 +34,5 @@ else
 
     mkdir -p artifacts
 
-    exec docker run -v $PWD:/v alpine:3.8 /v/packaging/alpine/$(basename $0) --in-docker
+    exec docker run -v $PWD:/v alpine:3.8 /v/packaging/alpine/$(basename $0) --in-docker $*
 fi
