@@ -40,7 +40,7 @@ extern "C" {
 #include "../src/rdkafka.h"
 }
 
-#ifdef _MSC_VER
+#ifdef _WIN32
 /* Visual Studio */
 #include "../src/win32_config.h"
 #else
@@ -101,6 +101,53 @@ void update_partitions_from_c_parts (std::vector<TopicPartition*> &partitions,
                                      const rd_kafka_topic_partition_list_t *c_parts);
 
 
+class ErrorImpl : public Error {
+ public:
+  ~ErrorImpl () {
+    rd_kafka_error_destroy(c_error_);
+  };
+
+  ErrorImpl (ErrorCode code, const std::string *errstr) {
+    c_error_ = rd_kafka_error_new(static_cast<rd_kafka_resp_err_t>(code),
+                                  errstr ? "%s" : NULL,
+                                  errstr ? errstr->c_str() : NULL);
+  }
+
+  ErrorImpl (rd_kafka_error_t *c_error):
+      c_error_(c_error) {};
+
+  static Error *create (ErrorCode code, const std::string *errstr) {
+    return new ErrorImpl(code, errstr);
+  }
+
+  ErrorCode   code () const {
+    return static_cast<ErrorCode>(rd_kafka_error_code(c_error_));
+  }
+
+  std::string name () const {
+    return std::string(rd_kafka_error_name(c_error_));
+  }
+
+  std::string str () const {
+    return std::string(rd_kafka_error_string(c_error_));
+  }
+
+  bool is_fatal () const {
+    return !!rd_kafka_error_is_fatal(c_error_);
+  }
+
+  bool is_retriable () const {
+    return !!rd_kafka_error_is_retriable(c_error_);
+  }
+
+  bool txn_requires_abort () const {
+    return !!rd_kafka_error_txn_requires_abort(c_error_);
+  }
+
+  rd_kafka_error_t *c_error_;
+};
+
+
 class EventImpl : public Event {
  public:
   ~EventImpl () {};
@@ -281,18 +328,20 @@ class MessageImpl : public Message {
       delete headers_;
   };
 
-  MessageImpl (RdKafka::Topic *topic, rd_kafka_message_t *rkmessage):
-  topic_(topic), rkmessage_(rkmessage), free_rkmessage_(true), key_(NULL),
-  headers_(NULL) {}
+  MessageImpl (rd_kafka_type_t rk_type,
+               RdKafka::Topic *topic, rd_kafka_message_t *rkmessage):
+      topic_(topic), rkmessage_(rkmessage),
+      free_rkmessage_(true), key_(NULL), headers_(NULL), rk_type_(rk_type) {}
 
-  MessageImpl (RdKafka::Topic *topic, rd_kafka_message_t *rkmessage,
+  MessageImpl (rd_kafka_type_t rk_type,
+               RdKafka::Topic *topic, rd_kafka_message_t *rkmessage,
                bool dofree):
-  topic_(topic), rkmessage_(rkmessage), free_rkmessage_(dofree), key_(NULL),
-  headers_(NULL) {}
+      topic_(topic), rkmessage_(rkmessage),
+      free_rkmessage_(dofree), key_(NULL), headers_(NULL), rk_type_(rk_type) {}
 
-  MessageImpl (rd_kafka_message_t *rkmessage):
-  topic_(NULL), rkmessage_(rkmessage), free_rkmessage_(true), key_(NULL),
-  headers_(NULL) {
+  MessageImpl (rd_kafka_type_t rk_type, rd_kafka_message_t *rkmessage):
+      topic_(NULL), rkmessage_(rkmessage),
+      free_rkmessage_(true), key_(NULL), headers_(NULL), rk_type_(rk_type)  {
     if (rkmessage->rkt) {
       /* Possibly NULL */
       topic_ = static_cast<Topic *>(rd_kafka_topic_opaque(rkmessage->rkt));
@@ -300,19 +349,20 @@ class MessageImpl : public Message {
   }
 
   /* Create errored message */
-  MessageImpl (RdKafka::Topic *topic, RdKafka::ErrorCode err):
-  topic_(topic), free_rkmessage_(false), key_(NULL), headers_(NULL) {
+  MessageImpl (rd_kafka_type_t rk_type,
+               RdKafka::Topic *topic, RdKafka::ErrorCode err):
+      topic_(topic), free_rkmessage_(false),
+      key_(NULL), headers_(NULL), rk_type_(rk_type)  {
     rkmessage_ = &rkmessage_err_;
     memset(&rkmessage_err_, 0, sizeof(rkmessage_err_));
     rkmessage_err_.err = static_cast<rd_kafka_resp_err_t>(err);
   }
 
   std::string         errstr() const {
-    /* FIXME: If there is an error string in payload (for consume_cb)
-     *        it wont be shown since 'payload' is reused for errstr
-     *        and we cant distinguish between consumer and producer.
-     *        For the producer case the payload needs to be the original
-     *        payload pointer. */
+    /* message_errstr() is only available for the consumer. */
+    if (rk_type_ == RD_KAFKA_CONSUMER)
+      return std::string(rd_kafka_message_errstr(rkmessage_));
+
     const char *es = rd_kafka_err2str(rkmessage_->err);
     return std::string(es ? es : "");
   }
@@ -390,6 +440,11 @@ class MessageImpl : public Message {
     return headers_;
   }
 
+  int32_t broker_id () const {
+    return rd_kafka_message_broker_id(rkmessage_);
+  }
+
+
   RdKafka::Topic *topic_;
   rd_kafka_message_t *rkmessage_;
   bool free_rkmessage_;
@@ -404,6 +459,7 @@ private:
   MessageImpl& operator=(MessageImpl const&) /*= delete*/;
 
   RdKafka::Headers *headers_;
+  const rd_kafka_type_t rk_type_; /**< Client type */
 };
 
 
@@ -828,7 +884,7 @@ class HandleImpl : virtual public Handle {
   int poll (int timeout_ms) { return rd_kafka_poll(rk_, timeout_ms); };
   int outq_len () { return rd_kafka_outq_len(rk_); };
 
-  void set_common_config (RdKafka::ConfImpl *confimpl);
+  void set_common_config (const RdKafka::ConfImpl *confimpl);
 
   RdKafka::ErrorCode metadata (bool all_topics,const Topic *only_rkt,
             Metadata **metadatap, int timeout_ms);
@@ -889,7 +945,7 @@ class HandleImpl : virtual public Handle {
           return rd_kafka_controllerid(rk_, timeout_ms);
   }
 
-  ErrorCode fatal_error (std::string &errstr) {
+  ErrorCode fatal_error (std::string &errstr) const {
           char errbuf[512];
           RdKafka::ErrorCode err =
                   static_cast<RdKafka::ErrorCode>(
@@ -1031,11 +1087,28 @@ public:
 };
 
 
+/**
+ * @class ConsumerGroupMetadata wraps the
+ *        C rd_kafka_consumer_group_metadata_t object.
+ */
+class ConsumerGroupMetadataImpl : public ConsumerGroupMetadata {
+ public:
+  ~ConsumerGroupMetadataImpl() {
+    rd_kafka_consumer_group_metadata_destroy(cgmetadata_);
+  }
+
+  ConsumerGroupMetadataImpl(rd_kafka_consumer_group_metadata_t *cgmetadata):
+      cgmetadata_(cgmetadata) {}
+
+  rd_kafka_consumer_group_metadata_t *cgmetadata_;
+};
+
 
 class KafkaConsumerImpl : virtual public KafkaConsumer, virtual public HandleImpl {
 public:
   ~KafkaConsumerImpl () {
-
+    if (rk_)
+      rd_kafka_destroy_flags(rk_, RD_KAFKA_DESTROY_F_NO_CONSUMER_CLOSE);
   }
 
   static KafkaConsumer *create (Conf *conf, std::string &errstr);
@@ -1107,6 +1180,16 @@ public:
   ErrorCode committed (std::vector<TopicPartition*> &partitions, int timeout_ms);
   ErrorCode position (std::vector<TopicPartition*> &partitions);
 
+  ConsumerGroupMetadata *groupMetadata () {
+    rd_kafka_consumer_group_metadata_t *cgmetadata;
+
+    cgmetadata = rd_kafka_consumer_group_metadata(rk_);
+    if (!cgmetadata)
+      return NULL;
+
+    return new ConsumerGroupMetadataImpl(cgmetadata);
+  }
+
   ErrorCode close ();
 
   ErrorCode seek (const TopicPartition &partition, int timeout_ms);
@@ -1198,7 +1281,10 @@ class ConsumerImpl : virtual public Consumer, virtual public HandleImpl {
 class ProducerImpl : virtual public Producer, virtual public HandleImpl {
 
  public:
-  ~ProducerImpl () { if (rk_) rd_kafka_destroy(rk_); };
+  ~ProducerImpl () {
+    if (rk_)
+      rd_kafka_destroy(rk_);
+  };
 
   ErrorCode produce (Topic *topic, int32_t partition,
                      int msgflags,
@@ -1241,6 +1327,71 @@ class ProducerImpl : virtual public Producer, virtual public HandleImpl {
                                                                 (int)purge_flags));
   }
 
+  Error *init_transactions (int timeout_ms) {
+    rd_kafka_error_t *c_error;
+
+    c_error = rd_kafka_init_transactions(rk_, timeout_ms);
+
+    if (c_error)
+      return new ErrorImpl(c_error);
+    else
+      return NULL;
+  }
+
+  Error *begin_transaction () {
+    rd_kafka_error_t *c_error;
+
+    c_error = rd_kafka_begin_transaction(rk_);
+
+    if (c_error)
+      return new ErrorImpl(c_error);
+    else
+      return NULL;
+  }
+
+  Error *send_offsets_to_transaction (
+      const std::vector<TopicPartition*> &offsets,
+      const ConsumerGroupMetadata *group_metadata,
+      int timeout_ms) {
+    rd_kafka_error_t *c_error;
+    const RdKafka::ConsumerGroupMetadataImpl *cgmdimpl =
+        dynamic_cast<const RdKafka::ConsumerGroupMetadataImpl *>(group_metadata);
+    rd_kafka_topic_partition_list_t *c_offsets = partitions_to_c_parts(offsets);
+
+    c_error = rd_kafka_send_offsets_to_transaction(rk_, c_offsets,
+                                                   cgmdimpl->cgmetadata_,
+                                                   timeout_ms);
+
+    rd_kafka_topic_partition_list_destroy(c_offsets);
+
+    if (c_error)
+      return new ErrorImpl(c_error);
+    else
+      return NULL;
+  }
+
+  Error *commit_transaction (int timeout_ms) {
+    rd_kafka_error_t *c_error;
+
+    c_error = rd_kafka_commit_transaction(rk_, timeout_ms);
+
+    if (c_error)
+      return new ErrorImpl(c_error);
+    else
+      return NULL;
+  }
+
+  Error *abort_transaction (int timeout_ms) {
+    rd_kafka_error_t *c_error;
+
+    c_error = rd_kafka_abort_transaction(rk_, timeout_ms);
+
+    if (c_error)
+      return new ErrorImpl(c_error);
+    else
+      return NULL;
+  }
+
   static Producer *create (Conf *conf, std::string &errstr);
 
 };
