@@ -29,12 +29,14 @@
 #ifndef _RDKAFKA_INT_H_
 #define _RDKAFKA_INT_H_
 
-#ifndef _MSC_VER
+#ifndef _WIN32
 #define _GNU_SOURCE  /* for strndup() */
-#include <syslog.h>
-#else
+#endif
+
+#ifdef _MSC_VER
 typedef int mode_t;
 #endif
+
 #include <fcntl.h>
 
 
@@ -56,10 +58,6 @@ typedef int mode_t;
 
 
 
-typedef struct rd_kafka_itopic_s rd_kafka_itopic_t;
-typedef struct rd_ikafka_s rd_ikafka_t;
-
-
 #define rd_kafka_assert(rk, cond) do {                                  \
                 if (unlikely(!(cond)))                                  \
                         rd_kafka_crash(__FILE__,__LINE__, __FUNCTION__, \
@@ -75,14 +73,12 @@ rd_kafka_crash (const char *file, int line, const char *function,
 
 /* Forward declarations */
 struct rd_kafka_s;
-struct rd_kafka_itopic_s;
+struct rd_kafka_topic_s;
 struct rd_kafka_msg_s;
 struct rd_kafka_broker_s;
 struct rd_kafka_toppar_s;
 
-typedef RD_SHARED_PTR_TYPE(, struct rd_kafka_toppar_s) shptr_rd_kafka_toppar_t;
-typedef RD_SHARED_PTR_TYPE(, struct rd_kafka_itopic_s) shptr_rd_kafka_itopic_t;
-
+typedef struct rd_kafka_lwtopic_s rd_kafka_lwtopic_t;
 
 
 #include "rdkafka_op.h"
@@ -97,6 +93,9 @@ typedef RD_SHARED_PTR_TYPE(, struct rd_kafka_itopic_s) shptr_rd_kafka_itopic_t;
 #include "rdkafka_assignor.h"
 #include "rdkafka_metadata.h"
 #include "rdkafka_mock.h"
+#include "rdkafka_partition.h"
+#include "rdkafka_coord.h"
+#include "rdkafka_mock.h"
 
 /**
  * Protocol level sanity
@@ -117,7 +116,10 @@ typedef RD_SHARED_PTR_TYPE(, struct rd_kafka_itopic_s) shptr_rd_kafka_itopic_t;
 typedef enum {
         RD_KAFKA_IDEMP_STATE_INIT,      /**< Initial state */
         RD_KAFKA_IDEMP_STATE_TERM,      /**< Instance is terminating */
+        RD_KAFKA_IDEMP_STATE_FATAL_ERROR, /**< A fatal error has been raised */
         RD_KAFKA_IDEMP_STATE_REQ_PID,   /**< Request new PID */
+        RD_KAFKA_IDEMP_STATE_WAIT_TRANSPORT, /**< Waiting for coordinator to
+                                              *   become available. */
         RD_KAFKA_IDEMP_STATE_WAIT_PID,  /**< PID requested, waiting for reply */
         RD_KAFKA_IDEMP_STATE_ASSIGNED,  /**< New PID assigned */
         RD_KAFKA_IDEMP_STATE_DRAIN_RESET, /**< Wait for outstanding
@@ -138,7 +140,9 @@ rd_kafka_idemp_state2str (rd_kafka_idemp_state_t state) {
         static const char *names[] = {
                 "Init",
                 "Terminate",
+                "FatalError",
                 "RequestPID",
+                "WaitTransport",
                 "WaitPID",
                 "Assigned",
                 "DrainReset",
@@ -150,12 +154,63 @@ rd_kafka_idemp_state2str (rd_kafka_idemp_state_t state) {
 
 
 
+/**
+ * @enum Transactional Producer state
+ */
+typedef enum {
+        /**< Initial state */
+        RD_KAFKA_TXN_STATE_INIT,
+        /**< Awaiting PID to be acquired by rdkafka_idempotence.c */
+        RD_KAFKA_TXN_STATE_WAIT_PID,
+        /**< PID acquired, but application has not made a successful
+         *   init_transactions() call. */
+        RD_KAFKA_TXN_STATE_READY_NOT_ACKED,
+        /**< PID acquired, no active transaction. */
+        RD_KAFKA_TXN_STATE_READY,
+        /**< begin_transaction() has been called. */
+        RD_KAFKA_TXN_STATE_IN_TRANSACTION,
+        /**< commit_transaction() has been called. */
+        RD_KAFKA_TXN_STATE_BEGIN_COMMIT,
+        /**< commit_transaction() has been called and all outstanding
+         *   messages, partitions, and offsets have been sent. */
+        RD_KAFKA_TXN_STATE_COMMITTING_TRANSACTION,
+        /**< abort_transaction() has been called. */
+        RD_KAFKA_TXN_STATE_ABORTING_TRANSACTION,
+        /**< An abortable error has occurred. */
+        RD_KAFKA_TXN_STATE_ABORTABLE_ERROR,
+        /* A fatal error has occured. */
+        RD_KAFKA_TXN_STATE_FATAL_ERROR
+} rd_kafka_txn_state_t;
+
+
+/**
+ * @returns the txn_state_t string representation
+ */
+static RD_UNUSED const char *
+rd_kafka_txn_state2str (rd_kafka_txn_state_t state) {
+        static const char *names[] = {
+                "Init",
+                "WaitPID",
+                "ReadyNotAcked",
+                "Ready",
+                "InTransaction",
+                "BeginCommit",
+                "CommittingTransaction",
+                "AbortingTransaction",
+                "AbortableError",
+                "FatalError"
+        };
+        return names[state];
+}
+
+
+
+
+
 /**
  * Kafka handle, internal representation of the application's rd_kafka_t.
  */
 
-typedef RD_SHARED_PTR_TYPE(shptr_rd_ikafka_s, rd_ikafka_t) shptr_rd_ikafka_t;
-
 struct rd_kafka_s {
 	rd_kafka_q_t *rk_rep;   /* kafka -> application reply queue */
 	rd_kafka_q_t *rk_ops;   /* any -> rdkafka main thread ops */
@@ -187,7 +242,7 @@ struct rd_kafka_s {
          * state changes. Protected by rk_broker_state_change_lock. */
         rd_list_t rk_broker_state_change_waiters; /**< (rd_kafka_enq_once_t*) */
 
-	TAILQ_HEAD(, rd_kafka_itopic_s)  rk_topics;
+	TAILQ_HEAD(, rd_kafka_topic_s)  rk_topics;
 	int              rk_topic_cnt;
 
         struct rd_kafka_cgrp_s *rk_cgrp;
@@ -198,7 +253,6 @@ struct rd_kafka_s {
 	rd_kafkap_str_t *rk_client_id;
         rd_kafkap_str_t *rk_group_id;    /* Consumer group id */
 
-	int              rk_flags;
 	rd_atomic32_t    rk_terminate;   /**< Set to RD_KAFKA_DESTROY_F_..
                                           *   flags instance
                                           *   is being destroyed.
@@ -274,6 +328,13 @@ struct rd_kafka_s {
         char            *rk_clusterid;      /* ClusterId from metadata */
         int32_t          rk_controllerid;   /* ControllerId from metadata */
 
+        /**< Producer: Delivery report mode */
+        enum {
+                RD_KAFKA_DR_MODE_NONE,  /**< No delivery reports */
+                RD_KAFKA_DR_MODE_CB,    /**< Delivery reports through callback */
+                RD_KAFKA_DR_MODE_EVENT, /**< Delivery reports through event API*/
+        } rk_drmode;
+
         /* Simple consumer count:
          *  >0: Running in legacy / Simple Consumer mode,
          *   0: No consumers running
@@ -286,6 +347,9 @@ struct rd_kafka_s {
          * @locks rk_lock
          */
         struct {
+                /*
+                 * Idempotence
+                 */
                 rd_kafka_idemp_state_t idemp_state; /**< Idempotent Producer
                                                      *   state */
                 rd_ts_t ts_idemp_state;/**< Last state change */
@@ -294,12 +358,151 @@ struct rd_kafka_s {
                 rd_atomic32_t inflight_toppar_cnt; /**< Current number of
                                                     *   toppars with inflight
                                                     *   requests. */
-                rd_kafka_timer_t request_pid_tmr; /**< Timer for pid retrieval*/
+                rd_kafka_timer_t pid_tmr; /**< PID FSM timer */
 
-                rd_kafkap_str_t *transactional_id; /**< Transactional Id,
-                                                    *   a null string. */
+                /*
+                 * Transactions
+                 *
+                 * All field access is from the rdkafka main thread,
+                 * unless a specific lock is mentioned in the doc string.
+                 *
+                 */
+                rd_atomic32_t txn_may_enq;      /**< Transaction state allows
+                                                 *   application to enqueue
+                                                 *   (produce) messages. */
+
+                rd_kafkap_str_t *transactional_id; /**< transactional.id */
+                rd_kafka_txn_state_t txn_state; /**< Transactional state.
+                                                 *   @locks rk_lock */
+                rd_ts_t ts_txn_state;           /**< Last state change.
+                                                 *   @locks rk_lock */
+                rd_kafka_broker_t *txn_coord;   /**< Transaction coordinator,
+                                                 *   this is a logical broker.*/
+                rd_kafka_broker_t *txn_curr_coord; /**< Current actual coord
+                                                    *   broker.
+                                                    *   This is only used to
+                                                    *   check if the coord
+                                                    *   changes. */
+                rd_kafka_broker_monitor_t txn_coord_mon; /**< Monitor for
+                                                          *   coordinator to
+                                                          *   take action when
+                                                          *   the broker state
+                                                          *   changes. */
+
+                /**< Blocking transactional API application call
+                 *   currently being handled, its state, reply queue and how
+                 *   to handle timeout.
+                 *   Only one transactional API call is allowed at any time.
+                 *   Protected by the rk_lock. */
+                struct {
+                        char name[64];        /**< API name, e.g.,
+                                               *   SendOffsetsToTransaction */
+                        rd_kafka_timer_t tmr; /**< Timeout timer, the timeout
+                                               * is specified by the app. */
+
+                        int flags;            /**< Flags */
+#define RD_KAFKA_TXN_CURR_API_F_ABORT_ON_TIMEOUT 0x1 /**< Set state to abortable
+                                                      *   error on timeout,
+                                                      *   i.e., fail the txn,
+                                                      *   and set txn_requires_abort
+                                                      *   on the returned error.
+                                                      */
+#define RD_KAFKA_TXN_CURR_API_F_RETRIABLE_ON_TIMEOUT 0x2 /**< Set retriable flag
+                                                          *   on the error
+                                                          *   on timeout. */
+#define RD_KAFKA_TXN_CURR_API_F_FOR_REUSE 0x4        /**< Do not reset the
+                                                      *   current API when it
+                                                      *   completes successfully
+                                                      *   Instead keep it alive
+                                                      *   and allow reuse with
+                                                      *   .._F_REUSE, blocking
+                                                      *   any non-F_REUSE
+                                                      *   curr API calls. */
+#define RD_KAFKA_TXN_CURR_API_F_REUSE     0x8        /**< Reuse/continue with
+                                                      *   current API state.
+                                                      *   This is used for
+                                                      *   multi-stage APIs,
+                                                      *   such as txn commit. */
+                } txn_curr_api;
+
+                /**< Copy (and reference) of the original init_transactions(),
+                 *   but out-lives the timeout of the curr API.
+                 *   This is used as the reply queue for when the
+                 *   black box idempotent producer has acquired the
+                 *   initial PID (or fails to do so).
+                 *   Since that acquisition may take longer than the
+                 *   init_transactions() API timeout this extra reference
+                 *   needs to be kept around.
+                 *   If the originating init_transactions() call has timed
+                 *   out and returned this queue reference simply points
+                 *   to a disabled queue that will discard any ops enqueued.
+                 *
+                 *   @locks rk_lock
+                 */
+                rd_kafka_q_t *txn_init_rkq;
+
+                int txn_req_cnt;                /**< Number of transaction
+                                                 *   requests sent.
+                                                 *   This is incremented when a
+                                                 *   AddPartitionsToTxn or
+                                                 *   AddOffsetsToTxn request
+                                                 *   has been sent for the
+                                                 *   current transaction,
+                                                 *   to keep track of
+                                                 *   whether the broker is
+                                                 *   aware of the current
+                                                 *   transaction and thus
+                                                 *   requires an EndTxn request
+                                                 *   on abort or not. */
+
+                /**< Timer to trigger registration of pending partitions */
+                rd_kafka_timer_t         txn_register_parts_tmr;
+
+                /**< Lock for txn_pending_rktps and txn_waitresp_rktps */
+                mtx_t                    txn_pending_lock;
+
+                /**< Partitions pending being added to transaction. */
+                rd_kafka_toppar_tqhead_t txn_pending_rktps;
+
+                /**< Partitions in-flight added to transaction. */
+                rd_kafka_toppar_tqhead_t txn_waitresp_rktps;
+
+                /**< Partitions added and registered to transaction. */
+                rd_kafka_toppar_tqhead_t txn_rktps;
+
+                /**< Number of messages that failed delivery.
+                 *   If this number is >0 on transaction_commit then an
+                 *   abortable transaction error will be raised.
+                 *   Is reset to zero on each begin_transaction(). */
+                rd_atomic64_t txn_dr_fails;
+
+                /**< Current transaction error. */
+                rd_kafka_resp_err_t txn_err;
+
+                /**< Current transaction error string, if any. */
+                char               *txn_errstr;
+
+                /**< Last InitProducerIdRequest error. */
+                rd_kafka_resp_err_t txn_init_err;
+
+                /**< Waiting for transaction coordinator query response */
+                rd_bool_t           txn_wait_coord;
+
+                /**< Transaction coordinator query timer */
+                rd_kafka_timer_t    txn_coord_tmr;
         } rk_eos;
 
+        /**<
+         * Coordinator cache.
+         *
+         * @locks none
+         * @locality rdkafka main thread
+         */
+        rd_kafka_coord_cache_t   rk_coord_cache; /**< Coordinator cache */
+
+        TAILQ_HEAD(, rd_kafka_coord_req_s) rk_coord_reqs; /**< Coordinator
+                                                           *   requests */
+
 	const rd_kafkap_bytes_t *rk_null_bytes;
 
 	struct {
@@ -364,6 +567,10 @@ struct rd_kafka_s {
                  *   but no more often than every 10s.
                  *   No locks: only accessed by rdkafka main thread. */
                 rd_interval_t broker_metadata_refresh;
+
+                /**< Suppression for allow.auto.create.topics=false not being
+                 *   supported by the broker. */
+                rd_interval_t allow_auto_create_topics;
         } rk_suppress;
 
         struct {
@@ -373,7 +580,16 @@ struct rd_kafka_s {
 
         /* Test mocks */
         struct {
-                rd_kafka_mock_cluster_t *cluster; /**< Mock cluster */
+                rd_kafka_mock_cluster_t *cluster; /**< Mock cluster, created
+                                                   *   by test.mock.num.brokers
+                                                   */
+                rd_atomic32_t cluster_cnt;        /**< Total number of mock
+                                                   *   clusters, created either
+                                                   *   through
+                                                   *   test.mock.num.brokers
+                                                   *   or mock_cluster_new().
+                                                   */
+
         } rk_mock;
 };
 
@@ -451,7 +667,8 @@ rd_kafka_curr_msgs_sub (rd_kafka_t *rk, unsigned int cnt, size_t size) {
 
         /* If the subtraction would pass one of the thresholds
          * broadcast a wake-up to any waiting listeners. */
-        if ((rk->rk_curr_msgs.cnt >= rk->rk_curr_msgs.max_cnt &&
+        if ((rk->rk_curr_msgs.cnt - cnt == 0) ||
+            (rk->rk_curr_msgs.cnt >= rk->rk_curr_msgs.max_cnt &&
              rk->rk_curr_msgs.cnt - cnt < rk->rk_curr_msgs.max_cnt) ||
             (rk->rk_curr_msgs.size >= rk->rk_curr_msgs.max_size &&
              rk->rk_curr_msgs.size - size < rk->rk_curr_msgs.max_size))
@@ -493,6 +710,25 @@ rd_kafka_curr_msgs_cnt (rd_kafka_t *rk) {
 	return cnt;
 }
 
+/**
+ * @brief Wait until \p tspec for curr_msgs to reach 0.
+ *
+ * @returns remaining curr_msgs
+ */
+static RD_INLINE RD_UNUSED int
+rd_kafka_curr_msgs_wait_zero (rd_kafka_t *rk, const struct timespec *tspec) {
+        int cnt;
+
+        mtx_lock(&rk->rk_curr_msgs.lock);
+        while ((cnt = rk->rk_curr_msgs.cnt) > 0) {
+                cnd_timedwait_abs(&rk->rk_curr_msgs.cnd,
+                                  &rk->rk_curr_msgs.lock,
+                                  tspec);
+        }
+        mtx_unlock(&rk->rk_curr_msgs.lock);
+
+        return cnt;
+}
 
 void rd_kafka_destroy_final (rd_kafka_t *rk);
 
@@ -533,7 +769,16 @@ int rd_kafka_simple_consumer_add (rd_kafka_t *rk);
  */
 #define rd_kafka_is_idempotent(rk) ((rk)->rk_conf.eos.idempotence)
 
-#define RD_KAFKA_PURGE_F_MASK 0x7
+/**
+ * @returns true if the producer is transactional (producer only).
+ */
+#define rd_kafka_is_transactional(rk)                   \
+        ((rk)->rk_conf.eos.transactional_id != NULL)
+
+
+#define RD_KAFKA_PURGE_F_ABORT_TXN 0x100  /**< Internal flag used when
+                                           *   aborting transaction */
+#define RD_KAFKA_PURGE_F_MASK 0x107
 const char *rd_kafka_purge_flags2str (int flags);
 
 
@@ -576,44 +821,54 @@ const char *rd_kafka_purge_flags2str (int flags);
 #define RD_KAFKA_DBG_ALL            0xfffff
 #define RD_KAFKA_DBG_NONE           0x0
 
+
 void rd_kafka_log0(const rd_kafka_conf_t *conf,
                    const rd_kafka_t *rk, const char *extra, int level,
+                   int ctx,
                    const char *fac, const char *fmt, ...) RD_FORMAT(printf,
-                                                                    6, 7);
+                                                                    7, 8);
 
-#define rd_kafka_log(rk,level,fac,...) \
-        rd_kafka_log0(&rk->rk_conf, rk, NULL, level, fac, __VA_ARGS__)
-#define rd_kafka_dbg(rk,ctx,fac,...) do {                               \
+#define rd_kafka_log(rk,level,fac,...)               \
+        rd_kafka_log0(&rk->rk_conf, rk, NULL, level, \
+                RD_KAFKA_DBG_NONE, fac, __VA_ARGS__)
+
+#define rd_kafka_dbg(rk,ctx,fac,...) do {                                   \
                 if (unlikely((rk)->rk_conf.debug & (RD_KAFKA_DBG_ ## ctx))) \
-                        rd_kafka_log0(&rk->rk_conf,rk,NULL,             \
-                                      LOG_DEBUG,fac,__VA_ARGS__);       \
+                        rd_kafka_log0(&rk->rk_conf,rk,NULL,                 \
+                                      LOG_DEBUG,(RD_KAFKA_DBG_ ## ctx),     \
+                                      fac,__VA_ARGS__);                     \
         } while (0)
 
 /* dbg() not requiring an rk, just the conf object, for early logging */
 #define rd_kafka_dbg0(conf,ctx,fac,...) do {                            \
                 if (unlikely((conf)->debug & (RD_KAFKA_DBG_ ## ctx)))   \
                         rd_kafka_log0(conf,NULL,NULL,                   \
-                                      LOG_DEBUG,fac,__VA_ARGS__);       \
+                                      LOG_DEBUG,(RD_KAFKA_DBG_ ## ctx), \
+                                      fac,__VA_ARGS__);                 \
         } while (0)
 
 /* NOTE: The local copy of _logname is needed due rkb_logname_lock lock-ordering
  *       when logging another broker's name in the message. */
-#define rd_rkb_log(rkb,level,fac,...) do {				\
-		char _logname[RD_KAFKA_NODENAME_SIZE];			\
-                mtx_lock(&(rkb)->rkb_logname_lock);                     \
+#define rd_rkb_log0(rkb,level,ctx,fac,...) do {                           \
+        char _logname[RD_KAFKA_NODENAME_SIZE];                            \
+                mtx_lock(&(rkb)->rkb_logname_lock);                       \
                 rd_strlcpy(_logname, rkb->rkb_logname, sizeof(_logname)); \
-                mtx_unlock(&(rkb)->rkb_logname_lock);                   \
-		rd_kafka_log0(&(rkb)->rkb_rk->rk_conf, \
-                              (rkb)->rkb_rk, _logname,                  \
-                              level, fac, __VA_ARGS__);                 \
+                mtx_unlock(&(rkb)->rkb_logname_lock);                     \
+        rd_kafka_log0(&(rkb)->rkb_rk->rk_conf,                            \
+                              (rkb)->rkb_rk, _logname,                    \
+                              level, ctx, fac, __VA_ARGS__);              \
         } while (0)
 
-#define rd_rkb_dbg(rkb,ctx,fac,...) do {				\
-		if (unlikely((rkb)->rkb_rk->rk_conf.debug &		\
-			     (RD_KAFKA_DBG_ ## ctx))) {			\
-			rd_rkb_log(rkb, LOG_DEBUG, fac, __VA_ARGS__);	\
-                }                                                       \
-	} while (0)
+#define rd_rkb_log(rkb,level,fac,...) \
+        rd_rkb_log0(rkb,level,RD_KAFKA_DBG_NONE,fac, __VA_ARGS__)
+
+#define rd_rkb_dbg(rkb,ctx,fac,...) do {                       \
+        if (unlikely((rkb)->rkb_rk->rk_conf.debug &            \
+                 (RD_KAFKA_DBG_ ## ctx))) {                    \
+            rd_rkb_log0(rkb, LOG_DEBUG,(RD_KAFKA_DBG_ ## ctx), \
+                    fac, __VA_ARGS__);                         \
+                }                                              \
+        } while (0)
 
 
 
@@ -637,29 +892,40 @@ rd_kafka_resp_err_t rd_kafka_set_last_error (rd_kafka_resp_err_t err,
 }
 
 
-int rd_kafka_set_fatal_error (rd_kafka_t *rk, rd_kafka_resp_err_t err,
-                              const char *fmt, ...) RD_FORMAT(printf, 3, 4);
+int rd_kafka_set_fatal_error0 (rd_kafka_t *rk, rd_dolock_t do_lock,
+                               rd_kafka_resp_err_t err,
+                               const char *fmt, ...) RD_FORMAT(printf, 4, 5);
+#define rd_kafka_set_fatal_error(rk,err,fmt,...)                        \
+        rd_kafka_set_fatal_error0(rk, RD_DO_LOCK, err, fmt, __VA_ARGS__)
 
 static RD_INLINE RD_UNUSED rd_kafka_resp_err_t
 rd_kafka_fatal_error_code (rd_kafka_t *rk) {
-        return rk->rk_conf.eos.idempotence &&
-                rd_atomic32_get(&rk->rk_fatal.err);
+        /* This is an optimization to avoid an atomic read which are costly
+         * on some platforms:
+         * Fatal errors are currently only raised by the idempotent producer
+         * and static consumers (group.instance.id). */
+        if ((rk->rk_type == RD_KAFKA_PRODUCER && rk->rk_conf.eos.idempotence) ||
+            (rk->rk_type == RD_KAFKA_CONSUMER && rk->rk_conf.group_instance_id))
+                return rd_atomic32_get(&rk->rk_fatal.err);
+
+        return RD_KAFKA_RESP_ERR_NO_ERROR;
 }
 
 
 extern rd_atomic32_t rd_kafka_thread_cnt_curr;
 extern char RD_TLS rd_kafka_thread_name[64];
 
-void rd_kafka_set_thread_name (const char *fmt, ...);
-void rd_kafka_set_thread_sysname (const char *fmt, ...);
+void rd_kafka_set_thread_name (const char *fmt, ...) RD_FORMAT(printf, 1, 2);
+void rd_kafka_set_thread_sysname (const char *fmt, ...) RD_FORMAT(printf, 1, 2);
 
 int rd_kafka_path_is_dir (const char *path);
+rd_bool_t rd_kafka_dir_is_empty (const char *path);
 
 rd_kafka_op_res_t
 rd_kafka_poll_cb (rd_kafka_t *rk, rd_kafka_q_t *rkq, rd_kafka_op_t *rko,
                   rd_kafka_q_cb_type_t cb_type, void *opaque);
 
-rd_kafka_resp_err_t rd_kafka_subscribe_rkt (rd_kafka_itopic_t *rkt);
+rd_kafka_resp_err_t rd_kafka_subscribe_rkt (rd_kafka_topic_t *rkt);
 
 
 /**
