@@ -34,7 +34,10 @@
 #include "rdkafka_interceptor.h"
 #include "rdkafka_header.h"
 #include "rdkafka_idempotence.h"
+#include "rdkafka_txnmgr.h"
+#include "rdkafka_error.h"
 #include "rdcrc32.h"
+#include "rdfnv1a.h"
 #include "rdmurmur2.h"
 #include "rdrand.h"
 #include "rdtime.h"
@@ -43,13 +46,72 @@
 
 #include <stdarg.h>
 
-void rd_kafka_msg_destroy (rd_kafka_t *rk, rd_kafka_msg_t *rkm) {
 
+const char *rd_kafka_message_errstr (const rd_kafka_message_t *rkmessage) {
+        if (!rkmessage->err)
+                return NULL;
+
+        if (rkmessage->payload)
+                return (const char *)rkmessage->payload;
+
+        return rd_kafka_err2str(rkmessage->err);
+}
+
+
+/**
+ * @brief Check if producing is allowed.
+ *
+ * @param errorp If non-NULL and an producing is prohibited a new error_t
+ *               object will be allocated and returned in this pointer.
+ *
+ * @returns an error if not allowed, else 0.
+ *
+ * @remarks Also sets the corresponding errno.
+ */
+static RD_INLINE rd_kafka_resp_err_t
+rd_kafka_check_produce (rd_kafka_t *rk, rd_kafka_error_t **errorp) {
+        rd_kafka_resp_err_t err;
+
+        if (unlikely((err = rd_kafka_fatal_error_code(rk)))) {
+                rd_kafka_set_last_error(RD_KAFKA_RESP_ERR__FATAL, ECANCELED);
+                if (errorp) {
+                        rd_kafka_rdlock(rk);
+                        *errorp = rd_kafka_error_new_fatal(
+                                err,
+                                "Producing not allowed since a previous fatal "
+                                "error was raised: %s",
+                                rk->rk_fatal.errstr);
+                        rd_kafka_rdunlock(rk);
+                }
+                return RD_KAFKA_RESP_ERR__FATAL;
+        }
+
+        if (likely(rd_kafka_txn_may_enq_msg(rk)))
+                return RD_KAFKA_RESP_ERR_NO_ERROR;
+
+        /* Transactional state forbids producing */
+        rd_kafka_set_last_error(RD_KAFKA_RESP_ERR__STATE, ENOEXEC);
+
+        if (errorp) {
+                rd_kafka_rdlock(rk);
+                *errorp = rd_kafka_error_new(
+                        RD_KAFKA_RESP_ERR__STATE,
+                        "Producing not allowed in transactional state %s",
+                        rd_kafka_txn_state2str(rk->rk_eos.txn_state));
+                rd_kafka_rdunlock(rk);
+        }
+
+        return RD_KAFKA_RESP_ERR__STATE;
+}
+
+
+void rd_kafka_msg_destroy (rd_kafka_t *rk, rd_kafka_msg_t *rkm) {
+//FIXME
 	if (rkm->rkm_flags & RD_KAFKA_MSG_F_ACCOUNT) {
 		rd_dassert(rk || rkm->rkm_rkmessage.rkt);
 		rd_kafka_curr_msgs_sub(
 			rk ? rk :
-			rd_kafka_topic_a2i(rkm->rkm_rkmessage.rkt)->rkt_rk,
+			rkm->rkm_rkmessage.rkt->rkt_rk,
 			1, rkm->rkm_len);
 	}
 
@@ -57,8 +119,7 @@ void rd_kafka_msg_destroy (rd_kafka_t *rk, rd_kafka_msg_t *rkm) {
                 rd_kafka_headers_destroy(rkm->rkm_headers);
 
 	if (likely(rkm->rkm_rkmessage.rkt != NULL))
-		rd_kafka_topic_destroy0(
-                        rd_kafka_topic_a2s(rkm->rkm_rkmessage.rkt));
+		rd_kafka_topic_destroy0(rkm->rkm_rkmessage.rkt);
 
 	if (rkm->rkm_flags & RD_KAFKA_MSG_F_FREE && rkm->rkm_payload)
 		rd_free(rkm->rkm_payload);
@@ -76,7 +137,7 @@ void rd_kafka_msg_destroy (rd_kafka_t *rk, rd_kafka_msg_t *rkm) {
  * @returns the new message
  */
 static
-rd_kafka_msg_t *rd_kafka_msg_new00 (rd_kafka_itopic_t *rkt,
+rd_kafka_msg_t *rd_kafka_msg_new00 (rd_kafka_topic_t *rkt,
 				    int32_t partition,
 				    int msgflags,
 				    char *payload, size_t len,
@@ -102,8 +163,9 @@ rd_kafka_msg_t *rd_kafka_msg_new00 (rd_kafka_itopic_t *rkt,
                                RD_KAFKA_MSG_F_FREE_RKM | msgflags);
 	rkm->rkm_len        = len;
 	rkm->rkm_opaque     = msg_opaque;
-	rkm->rkm_rkmessage.rkt = rd_kafka_topic_keep_a(rkt);
+	rkm->rkm_rkmessage.rkt = rd_kafka_topic_keep(rkt);
 
+        rkm->rkm_broker_id  = -1;
 	rkm->rkm_partition  = partition;
         rkm->rkm_offset     = RD_KAFKA_OFFSET_INVALID;
 	rkm->rkm_timestamp  = 0;
@@ -133,7 +195,6 @@ rd_kafka_msg_t *rd_kafka_msg_new00 (rd_kafka_itopic_t *rkt,
 		rkm->rkm_key_len = 0;
 	}
 
-
         return rkm;
 }
 
@@ -148,7 +209,7 @@ rd_kafka_msg_t *rd_kafka_msg_new00 (rd_kafka_itopic_t *rkt,
  * Returns 0 on success or -1 on error.
  * Both errno and 'errp' are set appropriately.
  */
-static rd_kafka_msg_t *rd_kafka_msg_new0 (rd_kafka_itopic_t *rkt,
+static rd_kafka_msg_t *rd_kafka_msg_new0 (rd_kafka_topic_t *rkt,
                                           int32_t force_partition,
                                           int msgflags,
                                           char *payload, size_t len,
@@ -238,7 +299,7 @@ static rd_kafka_msg_t *rd_kafka_msg_new0 (rd_kafka_itopic_t *rkt,
  *
  * @locks none
  */
-int rd_kafka_msg_new (rd_kafka_itopic_t *rkt, int32_t force_partition,
+int rd_kafka_msg_new (rd_kafka_topic_t *rkt, int32_t force_partition,
 		      int msgflags,
 		      char *payload, size_t len,
 		      const void *key, size_t keylen,
@@ -247,10 +308,8 @@ int rd_kafka_msg_new (rd_kafka_itopic_t *rkt, int32_t force_partition,
 	rd_kafka_resp_err_t err;
 	int errnox;
 
-        if (unlikely((err = rd_kafka_fatal_error_code(rkt->rkt_rk)))) {
-                rd_kafka_set_last_error(err, ECANCELED);
+        if (unlikely((err = rd_kafka_check_produce(rkt->rkt_rk, NULL))))
                 return -1;
-        }
 
         /* Create message */
         rkm = rd_kafka_msg_new0(rkt, force_partition, msgflags,
@@ -296,6 +355,188 @@ int rd_kafka_msg_new (rd_kafka_itopic_t *rkt, int32_t force_partition,
 }
 
 
+/** @remark Keep rd_kafka_produceva() and rd_kafka_producev() in synch */
+rd_kafka_error_t *rd_kafka_produceva (rd_kafka_t *rk,
+                                      const rd_kafka_vu_t *vus,
+                                      size_t cnt) {
+        rd_kafka_msg_t s_rkm = {
+                /* Message defaults */
+                .rkm_partition = RD_KAFKA_PARTITION_UA,
+                .rkm_timestamp = 0, /* current time */
+        };
+        rd_kafka_msg_t *rkm = &s_rkm;
+        rd_kafka_topic_t *rkt = NULL;
+        rd_kafka_resp_err_t err = RD_KAFKA_RESP_ERR_NO_ERROR;
+        rd_kafka_error_t *error = NULL;
+        rd_kafka_headers_t *hdrs = NULL;
+        rd_kafka_headers_t *app_hdrs = NULL; /* App-provided headers list */
+        size_t i;
+
+        if (unlikely(rd_kafka_check_produce(rk, &error)))
+                return error;
+
+        for (i = 0 ; i < cnt ; i++) {
+                const rd_kafka_vu_t *vu = &vus[i];
+                switch (vu->vtype)
+                {
+                case RD_KAFKA_VTYPE_TOPIC:
+                        rkt = rd_kafka_topic_new0(rk,
+                                                  vu->u.cstr,
+                                                  NULL, NULL, 1);
+                        break;
+
+                case RD_KAFKA_VTYPE_RKT:
+                        rkt = rd_kafka_topic_proper(vu->u.rkt);
+                        rd_kafka_topic_keep(rkt);
+                        break;
+
+                case RD_KAFKA_VTYPE_PARTITION:
+                        rkm->rkm_partition = vu->u.i32;
+                        break;
+
+                case RD_KAFKA_VTYPE_VALUE:
+                        rkm->rkm_payload = vu->u.mem.ptr;
+                        rkm->rkm_len = vu->u.mem.size;
+                        break;
+
+                case RD_KAFKA_VTYPE_KEY:
+                        rkm->rkm_key = vu->u.mem.ptr;
+                        rkm->rkm_key_len = vu->u.mem.size;
+                        break;
+
+                case RD_KAFKA_VTYPE_OPAQUE:
+                        rkm->rkm_opaque = vu->u.ptr;
+                        break;
+
+                case RD_KAFKA_VTYPE_MSGFLAGS:
+                        rkm->rkm_flags = vu->u.i;
+                        break;
+
+                case RD_KAFKA_VTYPE_TIMESTAMP:
+                        rkm->rkm_timestamp = vu->u.i64;
+                        break;
+
+                case RD_KAFKA_VTYPE_HEADER:
+                        if (unlikely(app_hdrs != NULL)) {
+                                error = rd_kafka_error_new(
+                                        RD_KAFKA_RESP_ERR__CONFLICT,
+                                        "VTYPE_HEADER and VTYPE_HEADERS "
+                                        "are mutually exclusive");
+                                goto err;
+                        }
+
+                        if (unlikely(!hdrs))
+                                hdrs = rd_kafka_headers_new(8);
+
+                        err = rd_kafka_header_add(hdrs,
+                                                  vu->u.header.name, -1,
+                                                  vu->u.header.val,
+                                                  vu->u.header.size);
+                        if (unlikely(err)) {
+                                error = rd_kafka_error_new(
+                                        err,
+                                        "Failed to add header: %s",
+                                        rd_kafka_err2str(err));
+                                goto err;
+                        }
+                        break;
+
+                case RD_KAFKA_VTYPE_HEADERS:
+                        if (unlikely(hdrs != NULL)) {
+                                error = rd_kafka_error_new(
+                                        RD_KAFKA_RESP_ERR__CONFLICT,
+                                        "VTYPE_HEADERS and VTYPE_HEADER "
+                                        "are mutually exclusive");
+                                goto err;
+                        }
+                        app_hdrs = vu->u.headers;
+                        break;
+
+                default:
+                        error = rd_kafka_error_new(
+                                RD_KAFKA_RESP_ERR__INVALID_ARG,
+                                "Unsupported VTYPE %d", (int)vu->vtype);
+                        goto err;
+                }
+        }
+
+        rd_assert(!error);
+
+        if (unlikely(!rkt)) {
+                error = rd_kafka_error_new(
+                        RD_KAFKA_RESP_ERR__INVALID_ARG,
+                        "Topic name or object required");
+                goto err;
+        }
+
+        rkm = rd_kafka_msg_new0(rkt,
+                                rkm->rkm_partition,
+                                rkm->rkm_flags,
+                                rkm->rkm_payload, rkm->rkm_len,
+                                rkm->rkm_key, rkm->rkm_key_len,
+                                rkm->rkm_opaque,
+                                &err, NULL,
+                                app_hdrs ? app_hdrs : hdrs,
+                                rkm->rkm_timestamp,
+                                rd_clock());
+
+        if (unlikely(err)) {
+                error = rd_kafka_error_new(
+                        err,
+                        "Failed to produce message: %s",
+                        rd_kafka_err2str(err));
+                goto err;
+        }
+
+        /* Partition the message */
+        err = rd_kafka_msg_partitioner(rkt, rkm, 1);
+        if (unlikely(err)) {
+                /* Handle partitioner failures: it only fails when
+                 * the application attempts to force a destination
+                 * partition that does not exist in the cluster. */
+
+                /* Interceptors: Unroll on_send by on_ack.. */
+                rkm->rkm_err = err;
+                rd_kafka_interceptors_on_acknowledgement(rk,
+                                                         &rkm->rkm_rkmessage);
+
+                /* Note we must clear the RD_KAFKA_MSG_F_FREE
+                 * flag since our contract says we don't free the payload on
+                 * failure. */
+                rkm->rkm_flags &= ~RD_KAFKA_MSG_F_FREE;
+
+                /* Deassociate application owned headers from message
+                 * since headers remain in application ownership
+                 * when producev() fails */
+                if (app_hdrs && app_hdrs == rkm->rkm_headers)
+                        rkm->rkm_headers = NULL;
+
+                rd_kafka_msg_destroy(rk, rkm);
+
+                error = rd_kafka_error_new(err,
+                                           "Failed to enqueue message: %s",
+                                           rd_kafka_err2str(err));
+                goto err;
+        }
+
+        rd_kafka_topic_destroy0(rkt);
+
+        return NULL;
+
+ err:
+        if (rkt)
+                rd_kafka_topic_destroy0(rkt);
+
+        if (hdrs)
+                rd_kafka_headers_destroy(hdrs);
+
+        rd_assert(error != NULL);
+        return error;
+}
+
+
+
+/** @remark Keep rd_kafka_produceva() and rd_kafka_producev() in synch */
 rd_kafka_resp_err_t rd_kafka_producev (rd_kafka_t *rk, ...) {
         va_list ap;
         rd_kafka_msg_t s_rkm = {
@@ -305,14 +546,12 @@ rd_kafka_resp_err_t rd_kafka_producev (rd_kafka_t *rk, ...) {
         };
         rd_kafka_msg_t *rkm = &s_rkm;
         rd_kafka_vtype_t vtype;
-        rd_kafka_topic_t *app_rkt;
-        shptr_rd_kafka_itopic_t *s_rkt = NULL;
-        rd_kafka_itopic_t *rkt;
+        rd_kafka_topic_t *rkt = NULL;
         rd_kafka_resp_err_t err;
         rd_kafka_headers_t *hdrs = NULL;
         rd_kafka_headers_t *app_hdrs = NULL; /* App-provided headers list */
 
-        if (unlikely((err = rd_kafka_fatal_error_code(rk))))
+        if (unlikely((err = rd_kafka_check_produce(rk, NULL))))
                 return err;
 
         va_start(ap, rk);
@@ -321,15 +560,15 @@ rd_kafka_resp_err_t rd_kafka_producev (rd_kafka_t *rk, ...) {
                 switch (vtype)
                 {
                 case RD_KAFKA_VTYPE_TOPIC:
-                        s_rkt = rd_kafka_topic_new0(rk,
-                                                    va_arg(ap, const char *),
-                                                    NULL, NULL, 1);
+                        rkt = rd_kafka_topic_new0(rk,
+                                                  va_arg(ap, const char *),
+                                                  NULL, NULL, 1);
                         break;
 
                 case RD_KAFKA_VTYPE_RKT:
-                        app_rkt = va_arg(ap, rd_kafka_topic_t *);
-                        s_rkt = rd_kafka_topic_keep(
-                                rd_kafka_topic_a2i(app_rkt));
+                        rkt = rd_kafka_topic_proper(
+                                va_arg(ap, rd_kafka_topic_t *));
+                        rd_kafka_topic_keep(rkt);
                         break;
 
                 case RD_KAFKA_VTYPE_PARTITION:
@@ -396,11 +635,9 @@ rd_kafka_resp_err_t rd_kafka_producev (rd_kafka_t *rk, ...) {
 
         va_end(ap);
 
-        if (unlikely(!s_rkt))
+        if (unlikely(!rkt))
                 return RD_KAFKA_RESP_ERR__INVALID_ARG;
 
-        rkt = rd_kafka_topic_s2i(s_rkt);
-
         if (likely(!err))
                 rkm = rd_kafka_msg_new0(rkt,
                                         rkm->rkm_partition,
@@ -414,7 +651,7 @@ rd_kafka_resp_err_t rd_kafka_producev (rd_kafka_t *rk, ...) {
                                         rd_clock());
 
         if (unlikely(err)) {
-                rd_kafka_topic_destroy0(s_rkt);
+                rd_kafka_topic_destroy0(rkt);
                 if (hdrs)
                         rd_kafka_headers_destroy(hdrs);
                 return err;
@@ -446,7 +683,7 @@ rd_kafka_resp_err_t rd_kafka_producev (rd_kafka_t *rk, ...) {
                 rd_kafka_msg_destroy(rk, rkm);
         }
 
-        rd_kafka_topic_destroy0(s_rkt);
+        rd_kafka_topic_destroy0(rkt);
 
         return err;
 }
@@ -463,7 +700,7 @@ int rd_kafka_produce (rd_kafka_topic_t *rkt, int32_t partition,
                       void *payload, size_t len,
                       const void *key, size_t keylen,
                       void *msg_opaque) {
-        return rd_kafka_msg_new(rd_kafka_topic_a2i(rkt), partition,
+        return rd_kafka_msg_new(rkt, partition,
                                 msgflags, payload, len,
                                 key, keylen, msg_opaque);
 }
@@ -486,17 +723,17 @@ int rd_kafka_produce_batch (rd_kafka_topic_t *app_rkt, int32_t partition,
         int multiple_partitions = (partition == RD_KAFKA_PARTITION_UA ||
                                    (msgflags & RD_KAFKA_MSG_F_PARTITION));
         rd_kafka_resp_err_t all_err;
-        rd_kafka_itopic_t *rkt = rd_kafka_topic_a2i(app_rkt);
-        shptr_rd_kafka_toppar_t *s_rktp = NULL;
+        rd_kafka_topic_t *rkt = rd_kafka_topic_proper(app_rkt);
+        rd_kafka_toppar_t *rktp = NULL;
 
         /* Propagated per-message below */
-        all_err = rd_kafka_fatal_error_code(rkt->rkt_rk);
+        all_err = rd_kafka_check_produce(rkt->rkt_rk, NULL);
 
         rd_kafka_topic_rdlock(rkt);
         if (!multiple_partitions) {
                 /* Single partition: look up the rktp once. */
-                s_rktp = rd_kafka_toppar_get_avail(rkt, partition,
-                                                   1/*ua on miss*/, &all_err);
+                rktp = rd_kafka_toppar_get_avail(rkt, partition,
+                                                 1/*ua on miss*/, &all_err);
 
         } else {
                 /* Indicate to lower-level msg_new..() that rkt is locked
@@ -544,24 +781,27 @@ int rd_kafka_produce_batch (rd_kafka_topic_t *app_rkt, int32_t partition,
                                         rd_kafka_msg_partitioner(
                                                 rkt, rkm, 0/*already locked*/);
                         } else {
-                                if (s_rktp == NULL ||
+                                if (rktp == NULL ||
                                     rkm->rkm_partition !=
-                                    rd_kafka_toppar_s2i(s_rktp)->
-                                    rktp_partition) {
+                                    rktp->rktp_partition) {
                                         rd_kafka_resp_err_t err;
-                                        if (s_rktp != NULL)
-                                                rd_kafka_toppar_destroy(s_rktp);
-                                        s_rktp = rd_kafka_toppar_get_avail(
+                                        if (rktp != NULL)
+                                                rd_kafka_toppar_destroy(rktp);
+                                        rktp = rd_kafka_toppar_get_avail(
                                                 rkt, rkm->rkm_partition,
                                                 1/*ua on miss*/, &err);
 
-                                        if (unlikely(!s_rktp)) {
+                                        if (unlikely(!rktp)) {
                                                 rkmessages[i].err = err;
                                                 continue;
                                         }
                                 }
-                                rd_kafka_toppar_enq_msg(
-                                        rd_kafka_toppar_s2i(s_rktp), rkm);
+                                rd_kafka_toppar_enq_msg(rktp, rkm);
+
+                                if (rd_kafka_is_transactional(rkt->rkt_rk)) {
+                                        /* Add partition to transaction */
+                                        rd_kafka_txn_add_partition(rktp);
+                                }
                         }
 
                         if (unlikely(rkmessages[i].err)) {
@@ -576,8 +816,7 @@ int rd_kafka_produce_batch (rd_kafka_topic_t *app_rkt, int32_t partition,
 
                 } else {
                         /* Single destination partition. */
-                        rd_kafka_toppar_enq_msg(rd_kafka_toppar_s2i(s_rktp),
-                                                rkm);
+                        rd_kafka_toppar_enq_msg(rktp, rkm);
                 }
 
                 rkmessages[i].err = RD_KAFKA_RESP_ERR_NO_ERROR;
@@ -585,8 +824,16 @@ int rd_kafka_produce_batch (rd_kafka_topic_t *app_rkt, int32_t partition,
         }
 
         rd_kafka_topic_rdunlock(rkt);
-        if (s_rktp != NULL)
-                rd_kafka_toppar_destroy(s_rktp);
+
+	if (!multiple_partitions && good > 0 &&
+            rd_kafka_is_transactional(rkt->rkt_rk) &&
+            rktp->rktp_partition != RD_KAFKA_PARTITION_UA) {
+                /* Add single destination partition to transaction */
+                rd_kafka_txn_add_partition(rktp);
+        }
+
+        if (rktp != NULL)
+                rd_kafka_toppar_destroy(rktp);
 
         return good;
 }
@@ -645,7 +892,7 @@ rd_kafka_msgq_enq_sorted0 (rd_kafka_msgq_t *rkmq,
         return ++rkmq->rkmq_msg_cnt;
 }
 
-int rd_kafka_msgq_enq_sorted (const rd_kafka_itopic_t *rkt,
+int rd_kafka_msgq_enq_sorted (const rd_kafka_topic_t *rkt,
                               rd_kafka_msgq_t *rkmq,
                               rd_kafka_msg_t *rkm) {
         rd_dassert(rkm->rkm_u.producer.msgid != 0);
@@ -745,16 +992,17 @@ void rd_kafka_msgq_split (rd_kafka_msgq_t *leftq, rd_kafka_msgq_t *rightq,
 /**
  * @brief Set per-message metadata for all messages in \p rkmq
  */
-void rd_kafka_msgq_set_metadata (rd_kafka_msgq_t *rkmq,
+void rd_kafka_msgq_set_metadata (rd_kafka_msgq_t *rkmq, int32_t broker_id,
                                  int64_t base_offset, int64_t timestamp,
                                  rd_kafka_msg_status_t status) {
         rd_kafka_msg_t *rkm;
 
         TAILQ_FOREACH(rkm, &rkmq->rkmq_msgs, rkm_link) {
+                rkm->rkm_broker_id = broker_id;
                 rkm->rkm_offset = base_offset++;
                 if (timestamp != -1) {
                         rkm->rkm_timestamp = timestamp;
-                        rkm->rkm_tstype = RD_KAFKA_MSG_ATTR_LOG_APPEND_TIME;
+                        rkm->rkm_tstype = RD_KAFKA_TIMESTAMP_LOG_APPEND_TIME;
                 }
 
                 /* Don't downgrade a message from any form of PERSISTED
@@ -860,17 +1108,46 @@ int32_t rd_kafka_msg_partitioner_murmur2_random (const rd_kafka_topic_t *rkt,
                 return (rd_murmur2(key, keylen) & 0x7fffffff) % partition_cnt;
 }
 
+int32_t rd_kafka_msg_partitioner_fnv1a (const rd_kafka_topic_t *rkt,
+                                        const void *key, size_t keylen,
+                                        int32_t partition_cnt,
+                                        void *rkt_opaque,
+                                        void *msg_opaque) {
+        return rd_fnv1a(key, keylen) % partition_cnt;
+}
+
+int32_t rd_kafka_msg_partitioner_fnv1a_random (const rd_kafka_topic_t *rkt,
+                                               const void *key, size_t keylen,
+                                               int32_t partition_cnt,
+                                               void *rkt_opaque,
+                                               void *msg_opaque) {
+        if (!key)
+                return rd_kafka_msg_partitioner_random(rkt,
+                                                       key,
+                                                       keylen,
+                                                       partition_cnt,
+                                                       rkt_opaque,
+                                                       msg_opaque);
+        else
+                return rd_fnv1a(key, keylen) % partition_cnt;
+}
+
 
 /**
- * Assigns a message to a topic partition using a partitioner.
- * Returns RD_KAFKA_RESP_ERR__UNKNOWN_PARTITION or .._UNKNOWN_TOPIC if
- * partitioning failed, or 0 on success.
+ * @brief Assigns a message to a topic partition using a partitioner.
+ *
+ * @param do_lock if RD_DO_LOCK then acquire topic lock.
+ *
+ * @returns RD_KAFKA_RESP_ERR__UNKNOWN_PARTITION or .._UNKNOWN_TOPIC if
+ *          partitioning failed, or 0 on success.
+ *
+ * @locality any
+ * @locks rd_kafka_
  */
-int rd_kafka_msg_partitioner (rd_kafka_itopic_t *rkt, rd_kafka_msg_t *rkm,
-			      int do_lock) {
+int rd_kafka_msg_partitioner (rd_kafka_topic_t *rkt, rd_kafka_msg_t *rkm,
+                              rd_dolock_t do_lock) {
 	int32_t partition;
 	rd_kafka_toppar_t *rktp_new;
-        shptr_rd_kafka_toppar_t *s_rktp_new;
 	rd_kafka_resp_err_t err;
 
 	if (do_lock)
@@ -893,6 +1170,14 @@ int rd_kafka_msg_partitioner (rd_kafka_itopic_t *rkt, rd_kafka_msg_t *rkm,
 			rd_kafka_topic_rdunlock(rkt);
                 return err;
 
+        case RD_KAFKA_TOPIC_S_ERROR:
+                /* Topic has permanent error.
+                 * Fail message immediately. */
+                err = rkt->rkt_err;
+                if (do_lock)
+                        rd_kafka_topic_rdunlock(rkt);
+                return err;
+
         case RD_KAFKA_TOPIC_S_EXISTS:
                 /* Topic exists in cluster. */
 
@@ -906,21 +1191,13 @@ int rd_kafka_msg_partitioner (rd_kafka_itopic_t *rkt, rd_kafka_msg_t *rkm,
 
                 /* Partition not assigned, run partitioner. */
                 if (rkm->rkm_partition == RD_KAFKA_PARTITION_UA) {
-                        rd_kafka_topic_t *app_rkt;
-                        /* Provide a temporary app_rkt instance to protect
-                         * from the case where the application decided to
-                         * destroy its topic object prior to delivery completion
-                         * (issue #502). */
-                        app_rkt = rd_kafka_topic_keep_a(rkt);
                         partition = rkt->rkt_conf.
-                                partitioner(app_rkt,
+                                partitioner(rkt,
                                             rkm->rkm_key,
 					    rkm->rkm_key_len,
                                             rkt->rkt_partition_cnt,
                                             rkt->rkt_conf.opaque,
                                             rkm->rkm_opaque);
-                        rd_kafka_topic_destroy0(
-                                rd_kafka_topic_a2s(app_rkt));
                 } else
                         partition = rkm->rkm_partition;
 
@@ -939,9 +1216,9 @@ int rd_kafka_msg_partitioner (rd_kafka_itopic_t *rkt, rd_kafka_msg_t *rkm,
         }
 
 	/* Get new partition */
-	s_rktp_new = rd_kafka_toppar_get(rkt, partition, 0);
+	rktp_new = rd_kafka_toppar_get(rkt, partition, 0);
 
-	if (unlikely(!s_rktp_new)) {
+	if (unlikely(!rktp_new)) {
 		/* Unknown topic or partition */
 		if (rkt->rkt_state == RD_KAFKA_TOPIC_S_NOTEXISTS)
 			err = RD_KAFKA_RESP_ERR__UNKNOWN_TOPIC;
@@ -954,7 +1231,6 @@ int rd_kafka_msg_partitioner (rd_kafka_itopic_t *rkt, rd_kafka_msg_t *rkm,
 		return  err;
 	}
 
-        rktp_new = rd_kafka_toppar_s2i(s_rktp_new);
         rd_atomic64_add(&rktp_new->rktp_c.producer_enq_msgs, 1);
 
         /* Update message partition */
@@ -965,7 +1241,14 @@ int rd_kafka_msg_partitioner (rd_kafka_itopic_t *rkt, rd_kafka_msg_t *rkm,
 	rd_kafka_toppar_enq_msg(rktp_new, rkm);
 	if (do_lock)
 		rd_kafka_topic_rdunlock(rkt);
-	rd_kafka_toppar_destroy(s_rktp_new); /* from _get() */
+
+        if (rktp_new->rktp_partition != RD_KAFKA_PARTITION_UA &&
+            rd_kafka_is_transactional(rkt->rkt_rk)) {
+                /* Add partition to transaction */
+                rd_kafka_txn_add_partition(rktp_new);
+        }
+
+	rd_kafka_toppar_destroy(rktp_new); /* from _get() */
 	return 0;
 }
 
@@ -989,6 +1272,8 @@ void rd_kafka_message_destroy (rd_kafka_message_t *rkmessage) {
 
 rd_kafka_message_t *rd_kafka_message_new (void) {
         rd_kafka_msg_t *rkm = rd_calloc(1, sizeof(*rkm));
+        rkm->rkm_flags      = RD_KAFKA_MSG_F_FREE_RKM;
+        rkm->rkm_broker_id  = -1;
         return (rd_kafka_message_t *)rkm;
 }
 
@@ -999,14 +1284,14 @@ rd_kafka_message_t *rd_kafka_message_new (void) {
  */
 static rd_kafka_message_t *
 rd_kafka_message_setup (rd_kafka_op_t *rko, rd_kafka_message_t *rkmessage) {
-        rd_kafka_itopic_t *rkt;
+        rd_kafka_topic_t *rkt;
         rd_kafka_toppar_t *rktp = NULL;
 
         if (rko->rko_type == RD_KAFKA_OP_DR) {
-                rkt = rd_kafka_topic_s2i(rko->rko_u.dr.s_rkt);
+                rkt = rko->rko_u.dr.rkt;
         } else {
                 if (rko->rko_rktp) {
-                        rktp = rd_kafka_toppar_s2i(rko->rko_rktp);
+                        rktp = rko->rko_rktp;
                         rkt = rktp->rktp_rkt;
                 } else
                         rkt = NULL;
@@ -1016,7 +1301,7 @@ rd_kafka_message_setup (rd_kafka_op_t *rko, rd_kafka_message_t *rkmessage) {
 
 
         if (!rkmessage->rkt && rkt)
-                rkmessage->rkt = rd_kafka_topic_keep_a(rkt);
+                rkmessage->rkt = rd_kafka_topic_keep(rkt);
 
         if (rktp)
                 rkmessage->partition = rktp->rktp_partition;
@@ -1121,6 +1406,15 @@ int64_t rd_kafka_message_latency (const rd_kafka_message_t *rkmessage) {
 }
 
 
+int32_t rd_kafka_message_broker_id (const rd_kafka_message_t *rkmessage) {
+        rd_kafka_msg_t *rkm;
+
+        rkm = rd_kafka_message2msg((rd_kafka_message_t *)rkmessage);
+
+        return rkm->rkm_broker_id;
+}
+
+
 
 /**
  * @brief Parse serialized message headers and populate
@@ -1288,9 +1582,9 @@ void rd_kafka_msgq_dump (FILE *fp, const char *what, rd_kafka_msgq_t *rkmq) {
  * @brief Destroy resources associated with msgbatch
  */
 void rd_kafka_msgbatch_destroy (rd_kafka_msgbatch_t *rkmb) {
-        if (rkmb->s_rktp) {
-                rd_kafka_toppar_destroy(rkmb->s_rktp);
-                rkmb->s_rktp = NULL;
+        if (rkmb->rktp) {
+                rd_kafka_toppar_destroy(rkmb->rktp);
+                rkmb->rktp = NULL;
         }
 
         rd_assert(RD_KAFKA_MSGQ_EMPTY(&rkmb->msgq));
@@ -1307,7 +1601,7 @@ void rd_kafka_msgbatch_init (rd_kafka_msgbatch_t *rkmb,
                              rd_kafka_pid_t pid) {
         memset(rkmb, 0, sizeof(*rkmb));
 
-        rkmb->s_rktp = rd_kafka_toppar_keep(rktp);
+        rkmb->rktp = rd_kafka_toppar_keep(rktp);
 
         rd_kafka_msgq_init(&rkmb->msgq);
 
@@ -1334,8 +1628,7 @@ void rd_kafka_msgbatch_set_first_msg (rd_kafka_msgbatch_t *rkmb,
          * need to handle wrapping. */
         rkmb->first_seq =
                 rd_kafka_seq_wrap(rkm->rkm_u.producer.msgid -
-                                  rd_kafka_toppar_s2i(rkmb->s_rktp)->
-                                  rktp_eos.epoch_base_msgid);
+                                  rkmb->rktp->rktp_eos.epoch_base_msgid);
 
         /* Check if there is a stored last message
          * on the first msg, which means an entire
@@ -1357,7 +1650,7 @@ void rd_kafka_msgbatch_set_first_msg (rd_kafka_msgbatch_t *rkmb,
  *         the toppar's in-flight count.
  */
 void rd_kafka_msgbatch_ready_produce (rd_kafka_msgbatch_t *rkmb) {
-        rd_kafka_toppar_t *rktp = rd_kafka_toppar_s2i(rkmb->s_rktp);
+        rd_kafka_toppar_t *rktp = rkmb->rktp;
         rd_kafka_t *rk = rktp->rktp_rkt->rkt_rk;
 
         /* Keep track of number of requests in-flight per partition,
@@ -1831,7 +2124,7 @@ unittest_msgq_insert_all_sort (const char *what,
         ut_rd_kafka_msgq_purge(&srcq);
         ut_rd_kafka_msgq_purge(&destq);
 
-        if (!rd_unittest_on_ci)
+        if (!rd_unittest_slow)
                 RD_UT_ASSERT(!(us_per_msg > max_us_per_msg + 0.0001),
                              "maximum us/msg exceeded: %.4f > %.4f us/msg",
                              us_per_msg, max_us_per_msg);
@@ -1941,7 +2234,7 @@ unittest_msgq_insert_each_sort (const char *what,
         RD_UT_SAY("Total: %.4fus/msg over %"PRId64" messages in %"PRId64"us",
                   us_per_msg, scnt, accum_ts);
 
-        if (!rd_unittest_on_ci)
+        if (!rd_unittest_slow)
                 RD_UT_ASSERT(!(us_per_msg > max_us_per_msg + 0.0001),
                              "maximum us/msg exceeded: %.4f > %.4f us/msg",
                              us_per_msg, max_us_per_msg);
