@@ -30,7 +30,7 @@ class LibrdkafkaTestApp(App):
 
         # Generate test config file
         conf_blob = list()
-        security_protocol='PLAINTEXT'
+        self.security_protocol='PLAINTEXT'
 
         f, self.test_conf_file = self.open_file('test.conf', 'perm')
         f.write('broker.address.family=v4\n'.encode('ascii'))
@@ -53,7 +53,7 @@ class LibrdkafkaTestApp(App):
         if mech != '':
             conf_blob.append('sasl.mechanisms=%s' % mech)
             if mech == 'PLAIN' or mech.find('SCRAM-') != -1:
-                security_protocol='SASL_PLAINTEXT'
+                self.security_protocol='SASL_PLAINTEXT'
                 # Use first user as SASL user/pass
                 for up in self.conf.get('sasl_users', '').split(','):
                     u,p = up.split('=')
@@ -62,12 +62,12 @@ class LibrdkafkaTestApp(App):
                     break
 
             elif mech == 'OAUTHBEARER':
-                security_protocol='SASL_PLAINTEXT'
+                self.security_protocol='SASL_PLAINTEXT'
                 conf_blob.append('enable.sasl.oauthbearer.unsecure.jwt=true\n')
                 conf_blob.append('sasl.oauthbearer.config=%s\n' % self.conf.get('sasl_oauthbearer_config'))
 
             elif mech == 'GSSAPI':
-                security_protocol='SASL_PLAINTEXT'
+                self.security_protocol='SASL_PLAINTEXT'
                 kdc = cluster.find_app(KerberosKdcApp)
                 if kdc is None:
                     self.log('WARNING: sasl_mechanisms is GSSAPI set but no KerberosKdcApp available: client SASL config will be invalid (which might be intentional)')
@@ -97,36 +97,36 @@ class LibrdkafkaTestApp(App):
 
             # Some tests need fine-grained access to various cert files,
             # set up the env vars accordingly.
-            for k, v in ssl.ca.iteritems():
+            for k, v in ssl.ca.items():
                 self.env_add('RDK_SSL_ca_{}'.format(k), v)
 
             # Set envs for all generated keys so tests can find them.
-            for k, v in key.iteritems():
+            for k, v in key.items():
                 if type(v) is dict:
-                    for k2, v2 in v.iteritems():
+                    for k2, v2 in v.items():
                         # E.g. "RDK_SSL_priv_der=path/to/librdkafka-priv.der"
                         self.env_add('RDK_SSL_{}_{}'.format(k, k2), v2)
                 else:
                     self.env_add('RDK_SSL_{}'.format(k), v)
 
 
-            if 'SASL' in security_protocol:
-                security_protocol = 'SASL_SSL'
+            if 'SASL' in self.security_protocol:
+                self.security_protocol = 'SASL_SSL'
             else:
-                security_protocol = 'SSL'
+                self.security_protocol = 'SSL'
 
         # Define bootstrap brokers based on selected security protocol
-        self.dbg('Using client security.protocol=%s' % security_protocol)
+        self.dbg('Using client security.protocol=%s' % self.security_protocol)
         all_listeners = (','.join(cluster.get_all('advertised.listeners', '', KafkaBrokerApp))).split(',')
-        bootstrap_servers = ','.join([x for x in all_listeners if x.startswith(security_protocol)])
+        bootstrap_servers = ','.join([x for x in all_listeners if x.startswith(self.security_protocol)])
         if len(bootstrap_servers) == 0:
             bootstrap_servers = all_listeners[0]
-            self.log('WARNING: No eligible listeners for security.protocol=%s in %s: falling back to first listener: %s: tests will fail (which might be the intention)' % (security_protocol, all_listeners, bootstrap_servers))
+            self.log('WARNING: No eligible listeners for security.protocol=%s in %s: falling back to first listener: %s: tests will fail (which might be the intention)' % (self.security_protocol, all_listeners, bootstrap_servers))
 
         self.bootstrap_servers = bootstrap_servers
 
         conf_blob.append('bootstrap.servers=%s' % bootstrap_servers)
-        conf_blob.append('security.protocol=%s' % security_protocol)
+        conf_blob.append('security.protocol=%s' % self.security_protocol)
 
         f.write(('\n'.join(conf_blob)).encode('ascii'))
         f.close()
@@ -150,7 +150,8 @@ class LibrdkafkaTestApp(App):
 
         # Per broker env vars
         for b in [x for x in self.cluster.apps if isinstance(x, KafkaBrokerApp)]:
-            self.env_add('BROKER_ADDRESS_%d' % b.appid, b.conf['address'])
+            self.env_add('BROKER_ADDRESS_%d' % b.appid,
+                         ','.join([x for x in b.conf['listeners'].split(',') if x.startswith(self.security_protocol)]))
             # Add each broker pid as an env so they can be killed indivdidually.
             self.env_add('BROKER_PID_%d' % b.appid, str(b.proc.pid))
             # JMX port, if available
