@@ -6,13 +6,13 @@ GREEN='\033[32m'
 CYAN='\033[36m'
 CCLR='\033[0m'
 
-if [ -z "$1" ]; then
+if [[ $1 == -h ]]; then
     echo "Usage: $0 [-..] [modes..]"
     echo ""
     echo "  Modes: bare valgrind helgrind drd gdb lldb bash"
     echo "  Options:"
     echo "   -..    - test-runner command arguments (pass thru)"
-    exit 1
+    exit 0
 fi
 
 ARGS=
@@ -61,6 +61,7 @@ for mode in $MODES; do
 	    valgrind $VALGRIND_ARGS --leak-check=full --show-leak-kinds=all \
 		     --errors-for-leak-kinds=all \
 		     --track-origins=yes \
+                     --track-fds=yes \
 		     $SUPP $GEN_SUPP \
 		$TEST $ARGS
 	    RET=$?
@@ -83,12 +84,18 @@ for mode in $MODES; do
 	    RET=$?
 	    ;;
         gdb)
-            if [[ -f gdb.run ]]; then
-                gdb -x gdb.run $ARGS $TEST
-            else
-                gdb $ARGS $TEST
-            fi
+            grun=$(mktemp gdbrunXXXXXX)
+            cat >$grun <<EOF
+set \$_exitcode = -999
+run $ARGS
+if \$_exitcode != -999
+ quit
+end
+EOF
+            export ASAN_OPTIONS="$ASAN_OPTIONS:abort_on_error=1"
+            gdb -x $grun $TEST
             RET=$?
+            rm $grun
             ;;
 	bare)
 	    $TEST $ARGS
